##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import zope.interface

class IExtrinsicReferences(zope.interface.Interface):
    """An object that stores extrinsic references to another object

    All objects must be adaptable to
    zope.app.keyreference.interfaces.IKeyReference."""

    def add(obj, value):
        """Add an object and an associated value to the registry.

        Both object and value must be adaptable to IKeyReference.

        Multiple values may be stored for a single key.  Each value is
        only stored once; comparisons are performed using the value's
        IKeyReference hash.
        """

    def update(obj, values):
        """For given object, add all values in iterable values.

        Object and each value must be adaptable to IKeyReference.  Identical
        values (as determined by IKeyReference) are collapsed to a single
        instance (so, for instance, a set of [A, B, B, C, B] will be
        collapsed to a logical set of A, B, C).
        """

    def get(obj):
        """Retrieve an iterable of the values associated with the object.

        If there are no references for `obj`, an iterable with no entries is
        returned.
        """

    def remove(obj, value):
        """Remove the specified value associated with the object.

        Comparisons are made with the IKeyReference hashes.

        If `value` is not set for `obj`, raises KeyError.
        """

    def contains(obj, value):
        """returns a boolean value of whether the obj : value pair exists."""

    def discard(obj, value):
        """Remove the specified value associated with the object.

        Comparisons are made with the IKeyReference hashes.

        If `value` is not set for `obj`, silently ignores.
        """

    def set(obj, values):
        """Set the values for obj to the values in the given iterable.

        Replaces any previous values for obj.  Object and each value must be
        adaptable to IKeyReference.  Identical values (as determined by
        IKeyReference) are collapsed to a single instance (so, for instance,
        values of [A, B, B, C, B] will be collapsed to a logical set of
        A, B, C).

        Setting an empty values is the canonical way of clearing values for an
        object.
        """



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Extrinsic references implementation"""

from zc.extrinsicreference.interfaces import IExtrinsicReferences
import BTrees
import persistent
import zope.app.keyreference.interfaces
import zope.component
import zope.interface


class ExtrinsicReferences(persistent.Persistent):

    zope.interface.implements(IExtrinsicReferences)

    # To be usable as an ILocalUtility we have to have these.
    __parent__ = __name__ = None

    def __init__(self):
        self.references = BTrees.OOBTree.OOBTree()

    def add(self, obj, value):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key)
        if refs is None:
            refs = self.references[key] = BTrees.OOBTree.OOTreeSet()
        refs.insert(zope.app.keyreference.interfaces.IKeyReference(value))

    def update(self, obj, values):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key)
        if refs is None:
            refs = self.references[key] = BTrees.OOBTree.OOTreeSet()
        refs.update(zope.app.keyreference.interfaces.IKeyReference(v)
            for v in values)

    def remove(self, obj, value):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key)
        if refs is not None:
            # raises KeyError when the value isn't found
            refs.remove(zope.app.keyreference.interfaces.IKeyReference(value))
        else:
            raise KeyError("Object and value pair does not exist")

    def discard(self, obj, value):
        try:
            self.remove(obj, value)
        except KeyError:
            pass

    def contains(self, obj, value):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key)
        if refs is not None:
            return zope.app.keyreference.interfaces.IKeyReference(value) in refs
        return False

    def set(self, obj, values):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key)
        vals = map(zope.app.keyreference.interfaces.IKeyReference, values)
        if not vals:
            if refs is not None:
                # del
                del self.references[key]
        else:
            if refs is None:
                refs = self.references[key] = BTrees.OOBTree.OOTreeSet()
            else:
                refs.clear()
            refs.update(vals)

    def get(self, obj):
        key = zope.app.keyreference.interfaces.IKeyReference(obj)
        refs = self.references.get(key, ())
        for kr in refs:
            yield kr()


def registerShortcut(shortcut, event):
    """Subscriber to add an extrinsic reference."""
    registry = zope.component.queryUtility(IExtrinsicReferences, 'shortcuts')
    if registry is not None:
        # We use raw_target because we don't want a proxy.
        registry.add(shortcut.raw_target, shortcut)

def unregisterShortcut(shortcut, event):
    """Subscriber to remove an extrinsic reference."""
    registry = zope.component.queryUtility(IExtrinsicReferences, 'shortcuts')
    if registry is not None:
        # We use raw_target because we don't want a proxy.
        registry.discard(shortcut.raw_target, shortcut)



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup for zc.extrinsicreference package
"""

import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name='zc.extrinsicreference',
    version='0.3.0',
    author='Zope Corporation and Contributors',
    author_email='zope-dev@zope.org',
    description='Extrinsic reference registries',
    long_description=(
        read('README.txt')
        + '\n\n' +
        'Detailed Documentation\n' +
        '----------------------'
        + '\n\n' +
        read('src', 'zc', 'extrinsicreference', 'README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        ),
    license='ZPL 2.1',
    keywords='extrinsic key reference zope zope3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Zope3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    url='http://pypi.python.org/pypi/zc.extrinsicreference',
    packages=find_packages('src'),
    package_dir={'':'src'},
    namespace_packages=['zc'],
    install_requires=[
        'ZODB3',
        'setuptools',
        'zc.shortcut',
        'zope.app.keyreference',
        'zope.component',
        'zope.interface',
        'zope.testing',
        ],
    include_package_data=True,
    zip_safe=False,
    )
